# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.
"""Band sieve.

This tool allows for performing the sieve of raster bands removing
patches having size lower than a threshold (i.e. number of pixels).

Typical usage example:

    >>> # import Remotior Sensus and start the session
    >>> import remotior_sensus
    >>> rs = remotior_sensus.Session()
    >>> # start the process
    >>> sieve = rs.band_sieve(input_bands=['file1.tif', 'file2.tif'],size=2,
    ... output_path='directory_path',connected=False,prefix='sieve_')
"""

from typing import Union, Optional

from remotior_sensus.core import configurations as cfg
from remotior_sensus.core.bandset_catalog import BandSet
from remotior_sensus.core.bandset_catalog import BandSetCatalog
from remotior_sensus.core.output_manager import OutputManager
from remotior_sensus.util import shared_tools


def band_sieve(
        input_bands: Union[list, int, BandSet], size: int,
        output_path: Union[list, str] = None, connected: Optional[bool] = None,
        overwrite: Optional[bool] = False,
        prefix: Optional[str] = '', extent_list: Optional[list] = None,
        n_processes: Optional[int] = None,
        available_ram: Optional[int] = None,
        bandset_catalog: Optional[BandSetCatalog] = None,
        virtual_output: Optional[bool] = None
) -> OutputManager:
    """Perform band sieve.

    This tool allows for performing the sieve of raster bands removing
    patches having size lower than a threshold (i.e. number of pixels).

    Args:
        input_bands: input of type BandSet or list of paths or integer
            number of BandSet.
        output_path: string of output path directory or list of paths.
        overwrite: if True, output overwrites existing files.
        size: size of dilation in pixels.
        virtual_output: if True (and output_path is directory), save output
            as virtual raster of multiprocess parts
        connected: if True, consider 8 pixel connection; if False, consider 4 pixel connection.
        prefix: optional string for output name prefix.
        extent_list: list of boundary coordinates left top right bottom.
        n_processes: number of parallel processes.
        available_ram: number of megabytes of RAM available to processes.
        bandset_catalog: optional type BandSetCatalog for BandSet number

    Returns:
        Object :func:`~remotior_sensus.core.output_manager.OutputManager` with
            - paths = output list
        
    Examples:
        Perform the sieve of size 3 with connected pixel (8 connection)
            >>> sieve = band_sieve(input_bands=['file1.tif', 'file2.tif'],size=3,output_path='directory_path',connected=True,prefix='sieve_')
    """  # noqa: E501
    cfg.logger.log.info('start')
    cfg.progress.update(
        process=__name__.split('.')[-1].replace('_', ' '), message='starting',
        start=True
    )
    # prepare process files
    prepared = shared_tools.prepare_process_files(
        input_bands=input_bands, output_path=output_path, overwrite=overwrite,
        n_processes=n_processes, bandset_catalog=bandset_catalog,
        box_coordinate_list=extent_list,
        prefix=prefix, multiple_output=True, multiple_input=True,
        virtual_output=virtual_output
    )
    input_raster_list = prepared['input_raster_list']
    raster_info = prepared['raster_info']
    n_processes = prepared['n_processes']
    nodata_list = prepared['nodata_list']
    output_list = prepared['output_list']
    # 4 connected pixels
    if connected:
        connected = 8
    elif not connected:
        connected = 4
    else:
        connected = 4
    # process calculation
    n = 0
    min_p = 1
    max_p = int((99 - 1) / len(input_raster_list))
    for i in input_raster_list:
        cfg.progress.update(message='processing raster %s' % (n + 1))
        out = output_list[n]
        nd = nodata_list[n]
        data_type = raster_info[n][8]
        # perform sieve
        cfg.multiprocess.multiprocess_raster_sieve(
            raster_path=i, n_processes=n_processes,
            available_ram=available_ram, sieve_size=size,
            connected=connected, output_nodata_value=nd, output=out,
            output_data_type=data_type, compress=cfg.raster_compression,
            min_progress=min_p + max_p * n,
            max_progress=min_p + max_p * (n + 1)
        )
        n += 1
    cfg.progress.update(end=True)
    cfg.logger.log.info('end; band sieve: %s' % output_list)
    return OutputManager(paths=output_list)



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.
"""Logging manger.

Core class that manages logs during processes.

Typical usage example:

    >>> # create a log file in a directory
    >>> Log(directory='directory_path', level=10)
"""

import io
import logging
from typing import Union, Optional

from remotior_sensus.core import configurations as cfg


class Log(object):
    log = None

    def __init__(
            self, file_path: Optional[str] = None,
            directory: Optional[str] = None, level: Union[int, str] = None,
            multiprocess=False, time=True
    ):
        """Manages logs.

        This module allows for managing logs of processes.

        Attributes:
            file_path: path of a log file.
            directory: directory path where a log file is created if file_path is None.
            level: level of logging (10 for DEBUG, 20 for INFO).
            multiprocess: if True, sets logging for parallel processes.
            time: if True, time is saved in log file.

        Examples:
            Create a log file and starts logging.
                >>> Log(file_path='file.txt', level=20)
        """  # noqa: E501
        if file_path is None:
            if directory is None:
                raise Exception('file path or directory missing')
            else:
                file_path = '{}/{}.log'.format(directory, cfg.root_name)
        # create logger
        logger = logging.getLogger(__name__)
        logging.basicConfig(level=logging.DEBUG)
        if level is None:
            level = logging.DEBUG
        if not multiprocess:
            # create file handler
            fh = logging.FileHandler(file_path)
            fh.setLevel(level)
            fhf = logging.Formatter(
                '%(levelname)s|%(asctime)s.%(msecs)03d|%(module)s|%(funcName)s'
                '|%(lineno)s|%(message)s', '%Y-%m-%dT%H:%M:%S'
            )
            fh.setFormatter(fhf)
            if logger.hasHandlers():
                logger.handlers.clear()
            logger.addHandler(fh)
            # create console handler
            ch = logging.StreamHandler()
            ch.setLevel(level)
            if time:
                chf = logging.Formatter(
                    '%(levelname)s[%(asctime)s.%(msecs)03d] '
                    '%(module)s.%(funcName)s[%(lineno)s] %(message)s',
                    '%Y-%m-%dT%H:%M:%S'
                )
            else:
                chf = logging.Formatter(
                    '%(levelname)s %(module)s.%(funcName)s[%(lineno)s] '
                    '%(message)s'
                )
            ch.setFormatter(chf)
            logger.addHandler(ch)
            logger.propagate = False
            self.log = logger
            self.file_path = file_path
            self.stream = None
            self.level = level
        # multiprocess number
        else:
            # create stream handler
            stream = io.StringIO()
            ch = logging.StreamHandler(stream)
            ch.setLevel(level)
            if time:
                chf = logging.Formatter(
                    '%(levelname)s_p{}|%(asctime)s.%(msecs)03d|%(module)s'
                    '|%(funcName)s|%(lineno)s|%(message)s'.format(
                        multiprocess
                    ), '%Y-%m-%dT%H:%M:%S'
                )
            else:
                chf = logging.Formatter(
                    '%(levelname)s_p{}|%(module)s|%(funcName)s|%(lineno)s'
                    '|%(message)s'.format(multiprocess)
                )
            ch.setFormatter(chf)
            if logger.hasHandlers():
                logger.handlers.clear()
            logger.addHandler(ch)
            logger.propagate = False
            self.log = logger
            self.stream = stream
            self.file_path = None
            self.level = level



from unittest import TestCase

import remotior_sensus
from remotior_sensus.util import files_directories


class TestBandErosion(TestCase):

    def test_band_erosion(self):
        rs = remotior_sensus.Session(
            n_processes=2, available_ram=1000, log_level=10
            )
        cfg = rs.configurations
        cfg.logger.log.debug('test')
        file_list = ['./data/S2_2020-01-01/S2_B02.tif',
                     './data/S2_2020-01-01/S2_B03.tif']
        cfg.logger.log.debug('>>> test band_erosion')
        erosion = rs.band_erosion(
            input_bands=file_list, output_path=cfg.temp.dir,
            value_list=[1, 425], size=1, circular_structure=True,
            prefix='erosion_'
            )
        self.assertTrue(files_directories.is_file(erosion.paths[0]))

        # clear temporary directory
        rs.close()



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.
"""
Band mask.

This tool allows for masking bands using a vector or raster mask.

Typical usage example:

    >>> # import Remotior Sensus and start the session
    >>> import remotior_sensus
    >>> rs = remotior_sensus.Session()
    >>> # start the process
    >>> rs.band_mask(input_bands=['path_1', 'path_2'], 
    ... input_mask='mask_path', mask_values=[1, 2]
    ... output_path='output_path')
"""  # noqa: E501

from typing import Union, Optional
import numpy as np

from remotior_sensus.core import configurations as cfg
from remotior_sensus.core.bandset_catalog import BandSet
from remotior_sensus.core.bandset_catalog import BandSetCatalog
from remotior_sensus.core.output_manager import OutputManager
from remotior_sensus.util import shared_tools, raster_vector
from remotior_sensus.tools.band_dilation import band_dilation
from remotior_sensus.core.processor_functions import band_calculation


def band_mask(
        input_bands: Union[list, int, BandSet],
        input_mask: str = None,
        mask_values: list = None,
        output_path: Optional[str] = None,
        overwrite: Optional[bool] = False,
        buffer: Optional[int] = None,
        nodata_value: Optional[int] = None,
        virtual_output: Optional[bool] = None,
        compress=None, compress_format=None,
        prefix: Optional[str] = '',
        extent_list: Optional[list] = None,
        n_processes: Optional[int] = None,
        available_ram: Optional[int] = None,
        bandset_catalog: Optional[BandSetCatalog] = None
):
    """Performs band mask.

    This tool allows for masking bands using a vector or raster mask.

    Args:
        input_bands: reference_raster of type BandSet or list of paths or integer
            number of BandSet.
        input_mask: string path of raster or vector used for masking.
        mask_values: list of values in the mask to be used for masking.
        output_path: string of output path directory or list of paths.
        overwrite: if True, output overwrites existing files.
        buffer: optional buffer size, in number of pixels, to expand the mask.
        nodata_value: value to be used as nodata in the output.
        virtual_output: if True (and output_path is directory), save output as virtual raster of multiprocess parts.
        prefix: optional string for output name prefix.
        extent_list: list of boundary coordinates left top right bottom.
        compress: if True, compress output.
        compress_format: format of compressions such as LZW or DEFLATE.
        n_processes: number of parallel processes.
        available_ram: number of megabytes of RAM available to processes.
        bandset_catalog: optional type BandSetCatalog for BandSet number.

    Returns:
        Object :func:`~remotior_sensus.core.output_manager.OutputManager` with
            - paths = output list

    Examples:
        Perform the band resample
            >>> band_mask(input_bands=['path_1', 'path_2'], 
            ... input_mask='mask_path', mask_values=[1, 2],
            ... output_path='output_path')
    """  # noqa: E501
    cfg.logger.log.info('start')
    cfg.progress.update(
        process=__name__.split('.')[-1].replace('_', ' '), message='starting',
        start=True
    )
    # prepare process files
    prepared = shared_tools.prepare_process_files(
        input_bands=input_bands, output_path=output_path, overwrite=overwrite,
        n_processes=n_processes, box_coordinate_list=extent_list,
        bandset_catalog=bandset_catalog, prefix=prefix,
        temporary_virtual_raster=True,
        multiple_output=True, virtual_output=virtual_output
    )
    input_raster_list = prepared['input_raster_list']
    raster_info = prepared['raster_info']
    n_processes = prepared['n_processes']
    nodata_list = prepared['nodata_list']
    output_list = prepared['output_list']
    # if vector convert to raster
    vector, raster, mask_crs = raster_vector.raster_or_vector_input(
        input_mask
    )
    # check crs
    same_crs = raster_vector.compare_crs(raster_info[0][1], mask_crs)
    # if reference is raster
    if raster:
        if not same_crs:
            t_pmd = cfg.temp.temporary_raster_path(extension=cfg.vrt_suffix)
            reference_raster = cfg.multiprocess.create_warped_vrt(
                raster_path=input_mask, output_path=t_pmd,
                output_wkt=str(mask_crs)
            )
        else:
            reference_raster = input_mask
    # if reference is vector
    else:
        if not same_crs:
            # project vector to raster crs
            t_vector = cfg.temp.temporary_file_path(
                name_suffix=cfg.gpkg_suffix
            )
            try:
                raster_vector.reproject_vector(
                    input_mask, t_vector, raster_info[0][1], mask_crs
                )
            except Exception as err:
                cfg.logger.log.error(str(err))
                cfg.messages.error(str(err))
                return OutputManager(check=False)
            input_mask = t_vector
        # convert vector to raster
        reference_raster = cfg.temp.temporary_raster_path(
            extension=cfg.tif_suffix
        )
        mask_values = [1]
        # perform conversion
        cfg.multiprocess.multiprocess_vector_to_raster(
            vector_path=input_mask,
            output_path=reference_raster, burn_values=1,
            reference_raster_path=input_raster_list[0], nodata_value=0,
            background_value=0, available_ram=available_ram,
            minimum_extent=False
        )
    if buffer is not None:
        size = int(buffer)
        vrt_file = cfg.temp.temporary_raster_path(extension=cfg.vrt_suffix)
        band_dilation(
            input_bands=[reference_raster], value_list=[1], size=size,
            output_path=[vrt_file], circular_structure=True
        )
        reference_raster = vrt_file
    cfg.logger.log.debug('reference_raster: %s' % reference_raster)
    argument_list = []
    scale_list = []
    offset_list = []
    output_nodata_list = []
    calculation_datatype = []
    output_datatype = []
    variables = []
    variable_list = []
    for b in range(0, len(input_raster_list)):
        if nodata_value is None:
            nodata_value = raster_info[b][4]
        output_nodata_list.append(nodata_value)
        output_datatype.append(raster_info[b][8])
        calculation_datatype.append(
            shared_tools.data_type_conversion(raster_info[b][8]))
        scale_list.append(raster_info[b][7][0])
        offset_list.append(raster_info[b][7][1])
        variables.append('"band%i"' % b)
        # function
        expression = ''
        closing = ''
        for c in mask_values:
            expression += 'np.where(%s[::, ::, 0] == %i, %i, ' % (
                cfg.array_function_placeholder, c, nodata_value)
            closing += ')'
        expression += '%s[::, ::, %i]%s' % (
            cfg.array_function_placeholder, b + 1, closing)
        argument_list.append(expression)
        variable_list.append(variables)
    cfg.logger.log.debug('argument_list: %s' % argument_list)
    # insert reference raster
    input_raster_list.insert(0, reference_raster)
    cfg.logger.log.debug('reference_raster: %s' % reference_raster)
    cfg.logger.log.debug('input_raster_list: %s' % input_raster_list)
    # prepare process files with reference
    prepared = shared_tools.prepare_process_files(
        input_bands=input_raster_list, output_path=output_path,
        overwrite=overwrite, n_processes=n_processes, prefix=prefix,
        temporary_virtual_raster=True,
        multiple_output=True, virtual_output=virtual_output
    )
    vrt_path_x = prepared['temporary_virtual_raster']
    # dummy bands for memory calculation
    dummy_bands = 2
    min_progress = 1
    one_progress = int((99 - 1) / (len(input_raster_list) - 1))
    max_progress = one_progress
    for band in range(len(input_raster_list) - 1):
        if output_datatype[band] == cfg.float64_dt:
            calc_data_type = np.float64
        elif output_datatype[band] == cfg.float32_dt:
            calc_data_type = np.float32
        elif (output_datatype[band] == cfg.int32_dt
              or output_datatype[band] == cfg.uint32_dt):
            calc_data_type = np.float32
        elif output_datatype[band] == cfg.int16_dt:
            calc_data_type = np.float32
        elif output_datatype[band] == cfg.uint16_dt:
            calc_data_type = np.float32
        elif output_datatype[band] == cfg.byte_dt:
            calc_data_type = np.float32
        else:
            calc_data_type = np.float32
        # run calculation
        cfg.multiprocess.run(
            raster_path=vrt_path_x, function=band_calculation,
            function_argument=argument_list[band],
            function_variable=variable_list[band],
            calculation_datatype=calc_data_type,
            use_value_as_nodata=nodata_list[band], dummy_bands=dummy_bands,
            output_raster_path=output_list[band],
            output_data_type=output_datatype[band],
            output_nodata_value=output_nodata_list[band],
            compress=compress, compress_format=compress_format,
            n_processes=n_processes, available_ram=available_ram,
            scale=scale_list[band], offset=offset_list[band],
            progress_message='masking',
            min_progress=min_progress, max_progress=max_progress
        )
        min_progress = max_progress
        max_progress += one_progress
    cfg.progress.update(end=True)
    cfg.logger.log.info('end; band resample: %s' % str(output_list))
    return OutputManager(paths=output_list)



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.

"""
Shared tools
"""

import re
from os import path
from random import randint
from typing import Union, Optional

import numpy as np

try:
    from scipy.ndimage import label
except Exception as error:
    str(error)

from remotior_sensus.core import configurations as cfg
from remotior_sensus.core.bandset_catalog import BandSet, BandSetCatalog
from remotior_sensus.util import files_directories, raster_vector
from remotior_sensus.core.processor_functions import region_growing


# check input path converting to the same crs if necessary
def prepare_input_list(
        band_list, reference_raster_crs=None,
        n_processes: int = None, src_nodata=None, dst_nodata=None
):
    cfg.logger.log.debug('start')
    cfg.logger.log.debug('band_list: %s' % str(band_list))
    information_list = []
    nodata_list = []
    if len(band_list) == 0:
        cfg.logger.log.error('empty list')
        return {'input_list': []}
    if reference_raster_crs is None:
        reference_raster_crs = raster_vector.get_crs(band_list[0])
    # get crs from file
    elif files_directories.is_file(reference_raster_crs):
        reference_raster_crs = raster_vector.get_crs(reference_raster_crs)
    input_list = []
    name_list = []
    gt_list = []
    xy_count_list = []
    box_coordinates_list = []
    warped = False
    for i in range(len(band_list)):
        name_list.append(files_directories.file_name(band_list[i]))
        crs = raster_vector.get_crs(band_list[i])
        # check crs
        same_crs = raster_vector.compare_crs(crs, reference_raster_crs)
        if not same_crs:
            t_pmd = cfg.temp.temporary_raster_path(
                name=files_directories.file_name(band_list[i]),
                extension=cfg.vrt_suffix
            )
            reference_raster = cfg.multiprocess.create_warped_vrt(
                raster_path=band_list[i], output_path=t_pmd,
                output_wkt=str(reference_raster_crs), n_processes=n_processes,
                src_nodata=src_nodata, dst_nodata=dst_nodata
            )
            warped = True
        else:
            reference_raster = band_list[i]
        input_list.append(reference_raster)
        info = raster_vector.raster_info(band_list[i])
        if info is not False:
            (gt, crs, un, xy_count, nd, number_of_bands, block_size,
             scale_offset, data_type) = info
        else:
            cfg.logger.log.error(
                'unable to get raster info: %s' % band_list[i]
            )
            return {'input_list': []}
        information_list.append(
            [gt, crs, un, xy_count, nd, number_of_bands, block_size,
             scale_offset, data_type]
        )
        gt_list.append(gt)
        xy_count_list.append(xy_count)
        nodata_list.append(nd)
        top = gt[3]
        left = gt[0]
        bottom = gt[3] + gt[5] * xy_count[1] + gt[4] * xy_count[0]
        right = gt[0] + gt[1] * xy_count[0] + gt[2] * xy_count[1]
        box_coordinates_list.append([left, top, right, bottom])
    check_gt_list = all(g_item == gt_list[0] for g_item in gt_list)
    check_xy_count_list = all(
        xy_item == xy_count_list[0] for xy_item in xy_count_list
    )
    if (check_gt_list is True and check_xy_count_list is True
            and warped is False):
        same_geotransformation = True
    else:
        same_geotransformation = False
    cfg.logger.log.debug('end; input_list: %s' % str(input_list))
    prepared = {
        'input_list': input_list, 'information_list': information_list,
        'nodata_list': nodata_list, 'name_list': name_list,
        'warped': warped, 'same_geotransformation': same_geotransformation,
        'box_coordinates_list': box_coordinates_list
    }
    return prepared


# prepare process files
def prepare_process_files(
        input_bands: Union[list, int, BandSet], output_path: str,
        n_processes: Optional[int] = None, overwrite=False,
        bandset_catalog: Optional[BandSetCatalog] = None,
        temporary_virtual_raster=True, prefix=None, multiple_output=False,
        multiple_input=False,
        virtual_output=None, box_coordinate_list: Optional[list] = None
):
    cfg.logger.log.debug('input_bands: %s' % str(input_bands))
    if n_processes is None:
        n_processes = cfg.n_processes
    # TODO check overwrite file
    if overwrite:
        pass
    # get input list
    band_list = BandSetCatalog.get_band_list(input_bands, bandset_catalog)
    if type(input_bands) is BandSet:
        coord_list = input_bands.box_coordinate_list
        if coord_list is not None:
            box_coordinate_list = coord_list
    elif type(input_bands) is int:
        coord_list = bandset_catalog.get_bandset(
            bandset_number=input_bands
        ).box_coordinate_list
        if coord_list is not None:
            box_coordinate_list = coord_list
    # list of inputs
    prepared_input = prepare_input_list(band_list, n_processes=n_processes)
    input_raster_list = prepared_input['input_list']
    if len(input_raster_list) == 0:
        return {'input_raster_list': []}
    raster_info = prepared_input['information_list']
    nodata_list = prepared_input['nodata_list']
    name_list = prepared_input['name_list']
    warped = prepared_input['warped']
    same_geotransformation = prepared_input['same_geotransformation']
    box_coordinates_list = prepared_input['box_coordinates_list']
    # single output path
    out_path = None
    # multiple output path list
    output_list = []
    # multiple output virtual raster check list
    vrt_list = []
    if multiple_output:
        if output_path is None:
            output_path = []
            for _r in input_raster_list:
                output_path.append(
                    cfg.temp.temporary_raster_path(
                        name=files_directories.file_name(_r),
                        extension=cfg.vrt_suffix
                    )
                )
        if type(output_path) is not list and files_directories.is_directory(
                output_path
        ):
            for r in input_raster_list:
                p = path.join(
                    output_path, '{}{}'.format(
                        prefix, files_directories.file_name(r)
                    )
                ).replace('\\', '/')
                # check output path
                out_path, vrt_r = files_directories.raster_output_path(
                    p, virtual_output
                )
                output_list.append(out_path)
                vrt_list.append(vrt_r)
        else:
            for r in output_path:
                # check output path
                out_path, vrt_r = files_directories.raster_output_path(
                    r, virtual_output
                )
                output_list.append(out_path)
                vrt_list.append(vrt_r)
    else:
        if output_path is None:
            output_path = cfg.temp.temporary_raster_path(
                extension=cfg.vrt_suffix
            )
        # check output path
        out_path, virtual_output = files_directories.raster_output_path(
            output_path, virtual_output
        )
    # create virtual raster of input
    if temporary_virtual_raster or box_coordinate_list is not None:
        # if rasters with same dimension don't create virtual raster but list
        if same_geotransformation is True and box_coordinate_list is None:
            temporary_virtual_raster = input_raster_list
        else:
            if multiple_input:
                temp_list = []
                if box_coordinate_list is None:
                    box_coordinate_list = box_coordinates_list[0]
                for r in input_raster_list:
                    temporary_virtual_raster = (
                        raster_vector.create_temporary_virtual_raster(
                            input_raster_list=[r],
                            box_coordinate_list=box_coordinate_list,
                            grid_reference=input_raster_list[0]
                        )
                    )
                    temp_list.append(temporary_virtual_raster)
                input_raster_list = temp_list
                temporary_virtual_raster = True
            else:
                temporary_virtual_raster = (
                    raster_vector.create_temporary_virtual_raster(
                        input_raster_list=input_raster_list,
                        box_coordinate_list=box_coordinate_list
                    )
                )
    prepared = {
        'input_raster_list': input_raster_list, 'raster_info': raster_info,
        'nodata_list': nodata_list, 'name_list': name_list, 'warped': warped,
        'output_path': out_path, 'virtual_output': virtual_output,
        'temporary_virtual_raster': temporary_virtual_raster,
        'n_processes': n_processes, 'output_list': output_list,
        'vrt_list': vrt_list, 'same_geotransformation': same_geotransformation
    }
    return prepared


# create base structure
def create_base_structure(size: int):
    structure = np.ones((size, size))
    return structure


# create a circular structure
def create_circular_structure(radius):
    circle = np.zeros([radius * 2 + 1, radius * 2 + 1])
    for x in range(radius * 2 + 1):
        for y in range(radius * 2 + 1):
            if (x - radius) ** 2 + (y - radius) ** 2 <= radius ** 2:
                circle[x, y] = 1
    return circle


# open structure file
def open_structure(structure):
    text = open(structure, 'r')
    lines = text.readlines()
    c = None
    for b in lines:
        b = b.replace('nan', 'np.nan').replace(cfg.new_line, '')
        a = eval(b)
        i = np.array(a, dtype=np.float32)
        try:
            c = np.append(c, [i], axis=0)
        except Exception as err:
            str(err)
            try:
                c = np.append([c], [i], axis=0)
            except Exception as err:
                str(err)
                c = i
    return c


# convert data type from string to numpy
def data_type_conversion(data_type):
    data_types = {
        cfg.float64_dt: np.float64, cfg.float32_dt: np.float32,
        cfg.int32_dt: np.int32, cfg.uint32_dt: np.uint32,
        cfg.int16_dt: np.int16, cfg.uint16_dt: np.uint16,
        cfg.byte_dt: np.uint8,
    }
    if data_type in data_types:
        return data_types[data_type]
    else:
        return None


# expand list of lists
def expand_list(input_list):
    expanded_list = []
    for c in input_list:
        expanded_list.extend(c)
    return expanded_list


# join path
def join_path(*argv):
    joined_path = path.join(*argv)
    return joined_path


# replace ignoring case
def replace(input_string: str, old_value: str, new_value: str):
    old_value = old_value.replace('*', '\\*').replace(
        '(', '\\('
    ).replace(')', '\\)')
    output = re.sub(old_value, new_value, input_string, flags=re.IGNORECASE)
    return output


# region growing polygon
def region_growing_polygon(
        coordinate_x, coordinate_y, input_bands: Union[list, int, BandSet],
        output_vector=None, band_number=None, max_width=None,
        max_spectral_distance=None, minimum_size=None, bandset_catalog=None,
        n_processes=None
):
    cfg.logger.log.debug('start')
    box_coordinate_list = tmp_vrt = None
    if type(input_bands) is BandSet:
        coord_list = input_bands.box_coordinate_list
        if coord_list is not None:
            box_coordinate_list = coord_list
    elif type(input_bands) is int:
        coord_list = bandset_catalog.get_bandset(
            bandset_number=input_bands
        ).box_coordinate_list
        if coord_list is not None:
            box_coordinate_list = coord_list
    # get input list
    band_list = BandSetCatalog.get_band_list(input_bands, bandset_catalog)
    cfg.logger.log.debug('band_list: %s' % str(band_list))
    # get input files
    prepared = prepare_input_list(
        band_list=band_list, n_processes=n_processes
    )
    input_raster_list = prepared['input_list']
    temporary_virtual_raster = raster_vector.create_temporary_virtual_raster(
        input_raster_list=input_raster_list,
        box_coordinate_list=box_coordinate_list
    )
    # input_path raster extent and pixel size
    info = raster_vector.image_geotransformation(temporary_virtual_raster)
    left = info['left']
    right = info['right']
    top = info['top']
    bottom = info['bottom']
    p_x = info['pixel_size_x']
    p_y = info['pixel_size_y']
    if max_width is None:
        max_width = 3
    if max_spectral_distance is None:
        max_spectral_distance = 1
    if minimum_size is None:
        minimum_size = 1
    # seed pixel
    abs_seed_x = abs(int((abs(coordinate_x) - left) / p_x))
    seed_x_min = left + abs_seed_x * p_x
    seed_x_max = left + (abs_seed_x + 1) * p_x
    abs_seed_y = abs(int((top - abs(coordinate_y)) / p_y))
    seed_y_max = top - abs_seed_y * p_y
    seed_y_min = top - (abs_seed_y + 1) * p_y
    region_min_x = seed_x_min - int(max_width / 2) * p_x
    if region_min_x < left:
        region_min_x = left
    region_max_x = seed_x_max + int(max_width / 2) * p_x
    if region_max_x > right:
        region_max_x = right
    region_min_y = seed_y_min - int(max_width / 2) * p_y
    if region_min_y < bottom:
        region_min_y = bottom
    region_max_y = seed_y_max + int(max_width / 2) * p_y
    if region_max_y > top:
        region_max_y = top
    seed_x = abs(int((abs(coordinate_x) - region_min_x) / p_x))
    seed_y = abs(int((region_max_y - abs(coordinate_y)) / p_y))
    cfg.logger.log.debug('seed_x: %s; seed_y: %s' % (seed_x, seed_y))
    # extract each virtual band and apply extent from max_width and coordinates
    extent_list = [region_min_x, region_max_y, region_max_x, region_min_y]
    function_variable = []
    virtual_raster_list = []
    if band_number is None:
        band_number_list = list(range(1, len(band_list) + 1))
        # extract virtual band
        for i in band_number_list:
            tmp_vrt = raster_vector.create_temporary_virtual_raster(
                input_raster_list=[temporary_virtual_raster],
                box_coordinate_list=extent_list,
                band_number_list=[[i]]
            )
            virtual_raster_list.append(tmp_vrt)
            function_variable.append(
                [seed_x, seed_y, float(max_spectral_distance),
                 int(minimum_size)]
            )
    else:
        band_number_list = [[band_number]]
        # extract virtual band
        tmp_vrt = raster_vector.create_temporary_virtual_raster(
            input_raster_list=[temporary_virtual_raster],
            box_coordinate_list=extent_list, band_number_list=band_number_list
        )
        virtual_raster_list.append(tmp_vrt)
        function_variable.append(
            [seed_x, seed_y, float(max_spectral_distance), int(minimum_size)]
        )
    cfg.multiprocess.run_separated(
        raster_path_list=virtual_raster_list, function=region_growing,
        function_argument=band_number_list,
        function_variable=function_variable, any_nodata_mask=True,
        keep_output_argument=True, n_processes=n_processes,
        progress_message='region growing'
    )
    cfg.multiprocess.multiprocess_region_growing()
    if cfg.multiprocess.output is False:
        cfg.logger.log.error('unable to calculate')
        return False
    regions = cfg.multiprocess.output
    if len(regions) > 0:
        region = None
        # run segmentation
        for region_array in regions:
            if region is None:
                region = region_array.astype(int)
            else:
                region = region * region_array.astype(int)
        if region[seed_y, seed_x] == 0:
            region[seed_y, seed_x] = 1
        if np.count_nonzero(region) > 0:
            try:
                region_label, num_features = label(region)
                region_seed_value = region_label[seed_y, seed_x]
                new_region = (region_label == region_seed_value).astype(int)
            except Exception as err:
                str(err)
                cfg.logger.log.error('region growing label')
                return False
        else:
            cfg.logger.log.error('region growing empty')
            return False
    else:
        new_region = regions[0]
    # create vector from region
    if np.count_nonzero(new_region) > 0 and new_region[seed_y, seed_x] == 1:
        if output_vector is None:
            output_vector = cfg.temp.temporary_file_path(
                name_suffix=cfg.gpkg_suffix
            )
        raster_vector.array_to_polygon(
            input_array=new_region, reference_raster=tmp_vrt,
            output=output_vector
        )
        cfg.logger.log.debug('end; roi vector: %s' % str(output_vector))
        return output_vector
    else:
        cfg.logger.log.error('region growing empty')
        return False


# calculate 2d histogram
def calculate_2d_histogram(x_values, y_values, decimal_round=None):
    if decimal_round is None:
        decimal_round = 3
    try:
        x_values_2 = x_values[~np.isnan(x_values)]
        y_values2 = y_values[~np.isnan(y_values)]
        x_values = x_values_2[~np.isnan(y_values2)]
        y_values = y_values2[~np.isnan(y_values2)]
        del x_values_2
        del y_values2
        x_max = np.amax(x_values)
        y_max = np.amax(y_values)
    except Exception as err:
        cfg.logger.log.error(str(err))
        return False
    precision = np.power(10, -float(decimal_round))
    x_min = np.amin(x_values)
    x_min = np.around(x_min, decimal_round) - precision
    x_max = np.around(x_max, decimal_round) + precision
    x_steps = calculate_steps(x_min, x_max, precision)
    y_min = np.amin(y_values)
    y_min = np.around(y_min, decimal_round) - precision
    y_max = np.around(y_max, decimal_round) + precision
    y_steps = calculate_steps(y_min, y_max, precision)
    try:
        histogram, x_edges, y_edges = np.histogram2d(
            x_values, y_values, bins=(x_steps, y_steps)
        )
    except Exception as err:
        cfg.logger.log.error(str(err))
        return None
    return [histogram, x_edges, y_edges]


# calculate steps
def calculate_steps(min_value, max_value, step):
    steps = [min_value]
    value = min_value
    while value < max_value:
        value += step
        steps.append(value)
    return steps


# get random color
def random_color():
    red = randint(0, 255)
    green = randint(0, 255)
    blue = randint(0, 255)
    return "#%02x%02x%02x" % (red, green, blue)


# calculate Bray-Curtis similarity
# (100 - 100 * sum(abs(x[ki]-x[kj]) / (sum(x[ki] + x[kj])))
def calculate_bray_curtis_similarity(values_x, values_y):
    try:
        array_i = np.array(values_x)
        array_j = np.array(values_y)
        sum_i_j = array_i.sum() + array_j.sum()
        absolute = np.sqrt((array_i - array_j) ** 2)
        value = 100 - absolute.sum() / sum_i_j * 100
    except Exception as err:
        str(err)
        value = cfg.not_available
    return value


# calculate Euclidean distance sqrt(sum((x[ki] - x[kj])^2))
def calculate_euclidean_distance(values_x, values_y):
    try:
        difference = (values_x - values_y) ** 2
        value = np.sqrt(difference.sum())
    except Exception as err:
        str(err)
        value = cfg.not_available
    return value


# calculate Spectral angle
# [ arccos( sum(r_i * s_i) / sqrt( sum(r_i**2) * sum(s_i**2) ) ) ]
def calculate_spectral_angle(values_x, values_y):
    try:
        value = np.arccos(
            (values_x * values_y).sum() / np.sqrt(
                (values_x ** 2).sum() * (values_y ** 2).sum()
            )
        ) * 180 / np.pi
    except Exception as err:
        str(err)
        value = cfg.not_available
    return value



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.

"""
PyTorch tools
"""

import numpy as np
from sklearn.model_selection import train_test_split

from remotior_sensus.core import configurations as cfg

nn_module = None
try:
    import torch
    from torch import nn
    nn_module = nn.Module
    from torch.utils.data import DataLoader, TensorDataset
except Exception as error:
    # empty class
    class Module:
        pass
    nn_module = Module
    if cfg.logger is not None:
        cfg.logger.log.error(str(error))


class PyTorchNeuralNetwork(nn_module):

    def __init__(
            self, columns, classes_number, hidden_layer_sizes=None,
            activation=None
    ):
        # call parent class
        super(PyTorchNeuralNetwork, self).__init__()
        if hidden_layer_sizes is not None:
            if activation == 'logistic' or activation == 'sigmoid':
                activation_function = nn.Sigmoid()
            elif activation == 'tanh':
                activation_function = nn.Tanh()
            elif activation == 'relu':
                activation_function = nn.ReLU()
            else:
                activation_function = nn.ReLU()
            layers = []
            for x in range(0, len(hidden_layer_sizes)):
                if x == 0:
                    layers.append(nn.Linear(columns, hidden_layer_sizes[x]))
                    layers.append(activation_function)
                else:
                    layers.append(
                        nn.Linear(
                            hidden_layer_sizes[x - 1], hidden_layer_sizes[x]
                        )
                    )
                    layers.append(activation_function)
            layers.append(
                nn.Linear(hidden_layer_sizes[-1], int(classes_number))
            )
            self.sequential = nn.Sequential(*layers)
        else:
            self.sequential = nn.Sequential(
                nn.Linear(columns, 100), nn.ReLU(),
                nn.Linear(100, classes_number)
            )

    # forward
    def forward(self, training_data):
        logits = self.sequential(training_data)
        return logits


# noinspection PyTypeChecker,PyUnresolvedReferences
def train_pytorch_model(
        x_matrix, y_matrix, pytorch_model=None, activation='relu',
        batch_size=None, n_processes=0, training_portion=None,
        pytorch_optimizer=None, hidden_layer_sizes=None,
        loss_function=None, learning_rate_init=None,
        optimization_n_iter_no_change=None, optimization_tol=None,
        weight_decay=None, max_iterations=None, device=None, min_progress=0,
        max_progress=100
):
    cfg.logger.log.debug('start')
    # get device
    if device == 'cuda':
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
    else:
        device = 'cpu'
    if device == 'cpu':
        torch.set_num_threads(n_processes)
    if optimization_n_iter_no_change is None:
        optimization_n_iter_no_change = 5
    if optimization_tol is None:
        optimization_tol = 0.0001
    if max_iterations is None:
        max_iterations = 200
    if batch_size is None or batch_size == 'auto':
        batch_size = int(
            min(
                2000, min(x_matrix.shape[0], max(200, x_matrix.shape[0] / 100))
            )
        )
    cfg.logger.log.debug(
        'device: %s; n_processes: %s; batch_size: %s' % (
            device, n_processes, batch_size)
    )
    # number of classes
    classes_number = np.max(y_matrix) + 1
    if pytorch_model is None:
        # move model to device
        model = PyTorchNeuralNetwork(
            columns=x_matrix.shape[1], classes_number=classes_number,
            hidden_layer_sizes=hidden_layer_sizes, activation=activation
        ).to(device)
    else:
        model = pytorch_model(
            columns=x_matrix.shape[1], classes_number=classes_number,
            hidden_layer_sizes=hidden_layer_sizes, activation=activation
        ).to(device)
    data_type = eval('torch.%s' % x_matrix.dtype)
    if training_portion is not None and training_portion < 1:
        x_train, x_test, y_train, y_test = train_test_split(
            x_matrix, y_matrix, test_size=1 - training_portion,
            train_size=training_portion, random_state=0,
            stratify=y_matrix
        )
        x_train = torch.tensor(x_train, dtype=data_type)
        y_train = torch.tensor(y_train, dtype=torch.long)
        x_test = torch.tensor(x_test, dtype=data_type)
        y_test = torch.tensor(y_test, dtype=torch.long)
        training_data = DataLoader(
            TensorDataset(x_train, y_train), batch_size=batch_size,
            shuffle=True
        )
        test_data = DataLoader(
            TensorDataset(x_test, y_test), batch_size=batch_size, shuffle=True
        )
    else:
        x_train, x_test, y_train, y_test = train_test_split(
            x_matrix, y_matrix, test_size=0.1,
            train_size=0.9, random_state=0, stratify=y_matrix
        )
        x_train = torch.tensor(x_matrix, dtype=data_type)
        y_train = torch.tensor(y_matrix, dtype=torch.long)
        x_test = torch.tensor(x_test, dtype=data_type)
        y_test = torch.tensor(y_test, dtype=torch.long)
        training_data = DataLoader(
            TensorDataset(x_train, y_train), batch_size=batch_size,
            shuffle=True
        )
        test_data = DataLoader(
            TensorDataset(x_test, y_test), batch_size=batch_size, shuffle=True
        )
    if learning_rate_init is None:
        learning_rate_init = 0.001
    if weight_decay is None:
        weight_decay = 0.0001
    # optimizer
    if pytorch_optimizer is None:
        pytorch_optimizer = torch.optim.Adam(
            model.parameters(), lr=learning_rate_init,
            weight_decay=weight_decay
        )
    # loss function
    if loss_function is None:
        loss_function = nn.CrossEntropyLoss()
    epoch = 0
    test_loss_list = []
    while True:
        if cfg.action is True:
            epoch += 1
            # train
            model.train()
            training_size = len(training_data.dataset)
            training_loss = 0
            for batch, (x, y) in enumerate(training_data):
                x, y = x.to(device), y.to(device)
                # compute prediction loss
                prediction = model(x)
                loss = loss_function(prediction, y)
                # backpropagation
                pytorch_optimizer.zero_grad()
                loss.backward()
                pytorch_optimizer.step()
                training_loss += loss.item()
                # progress
                count_progress = batch * len(x)
                cfg.progress.update(
                    percentage=int(100 * count_progress / training_size)
                )
            training_loss /= training_size
            # test
            model.eval()
            test_size = len(test_data.dataset)
            test_loss, correct = 0, 0
            with torch.no_grad():
                for x, y in test_data:
                    x, y = x.to(device), y.to(device)
                    prediction = model(x)
                    test_loss += loss_function(prediction, y).item()
                    correct += (prediction.argmax(1) == y).type(
                        torch.float
                    ).sum().item()
            test_loss /= test_size
            test_loss_list.append(test_loss)
            accuracy = correct / test_size * 100
            # progress
            if max_iterations is not None:
                increment = (max_progress - min_progress) / max_iterations
                step = int(min_progress + epoch * increment)
            else:
                step = None
            cfg.progress.update(
                message=f'epoch: {epoch}, training loss: {training_loss:>8f}, '
                        f'test loss: {test_loss:>8f}, '
                        f'accuracy: {accuracy :>0.1f}%', step=step
            )
            cfg.logger.log.debug(
                f'epoch: {epoch}, training loss: {training_loss:>8f}, '
                f'test loss: {test_loss:>8f}, '
                f'accuracy: {accuracy :>0.1f}%'
            )
            # check optimization tolerance
            if epoch > optimization_n_iter_no_change:
                loss_difference = []
                for o in range(1, optimization_n_iter_no_change + 1):
                    diff = test_loss_list[-1 * o] - test_loss_list[
                        -1 * (o + 1)]
                    loss_difference.append(diff ** 2 < optimization_tol ** 2)
                if all(loss_difference):
                    cfg.logger.log.debug(
                        'optimization_tol: %s' % str(optimization_tol)
                    )
                    break
            if max_iterations is not None and epoch == max_iterations:
                cfg.logger.log.debug('max_iterations: %s'
                                     % str(max_iterations))
                break
        else:
            cfg.logger.log.error('cancel')
            return None, None, None, None
    cfg.logger.log.debug('end')
    return model, training_loss, test_loss, accuracy
